import pytest
import logging

from assertpy import assert_that
import os

from netdot import Repository, exceptions
import netdot


@pytest.mark.vcr
def test_show_changes(repository: Repository, capfd):
    # Arrange
    repository.enable_propose_changes(print_changes=False)
    repository.create_new(netdot.Audit())
    repository.create_new(netdot.Availability())
    repository.create_new(netdot.HorizontalCable())
    site = repository.get_site(137)
    site.name = 'AAAA'
    repository.create_new(netdot.Device())
    repository.delete(netdot.BGPPeering())
    repository.create_new(netdot.BGPPeering())
    site.aliases = 'BBBB'  # Later, make some more updates to the site

    # Act
    repository.show_changes(terse=True)

    # Assert
    console_output = capfd.readouterr().out
    assert_that(console_output).contains(' 1. Will CREATE Audit: Audit(id=None, fields=None, label=None, object_id=None...')
    assert_that(console_output).contains('2. Will CREATE Availability')
    assert_that(console_output).contains('3. Will CREATE HorizontalCable')
    assert_that(console_output).contains('4. Will CREATE Device')
    assert_that(console_output).contains('5. Will DELETE BGPPeering')
    assert_that(console_output).contains('6. Will CREATE BGPPeering')
    assert_that(console_output).matches('7. Will UPDATE Site.*AAAA.*BBBB.*')


@pytest.mark.vcr
def test_get_rr_by_address_does_NOT_propose_changes(repository: Repository, capfd):
    # Arrange
    repository.enable_propose_changes(print_changes=False)
    repository.get_rr_by_address('128.223.250.151')

    # Act
    repository.show_changes()

    # Assert
    console_output = capfd.readouterr().out
    assert_that(console_output.strip()).is_equal_to('None, yet...')


@pytest.mark.vcr
def test_show_changes_as_tables(repository: Repository, capfd):
    # Arrange
    repository.enable_propose_changes(print_changes=False)
    repository.create_new(netdot.Audit())
    repository.create_new(netdot.Device(collect_arp=False))
    repository.delete(netdot.BGPPeering())
    site = repository.get_site(137)
    site.name = 'UPDATED'
    site.aliases = 'AAAA'

    # Act
    repository.show_changes_as_tables(terse=True)

    # Assert
    console_output = capfd.readouterr().out
    assert_that(console_output).contains('''## BGPPeering Changes

action    id    bgppeeraddr    bgppeerid    device_xlink    entity_xlink
--------  ----  -------------  -----------  --------------  --------------
DELETE    None  None           None         None            None
''')
    assert_that(console_output).contains('''## Device Changes

action    id    site_xlink    asset_id_xlink    monitorstatus_xlink    name_xlink
--------  ----  ------------  ----------------  ---------------------  ------------
CREATE    None  None          None              None                   None
''')
    assert_that(console_output).contains('''## Site Changes

action      id  name              aliases    availability_xlink      contactlist_xlink
--------  ----  ----------------  ---------  --------------------  -------------------
UPDATE     137  [-Computing       +AAAA+     None                                  144
                Center
                (039)-]+UPDATED+''')


@pytest.mark.vcr
def test_show_changes_as_tables_widescreen(repository: Repository, capfd):
    # Arrange
    repository.enable_propose_changes(print_changes=False)
    site = repository.get_site(137)
    site.name = 'AAAA'
    site.aliases = 'BBBB'

    # Act
    repository.show_changes_as_tables(terse=False)

    # Assert
    console_output = capfd.readouterr().out
    assert_that(console_output).contains('''## Site Changes

action      id  name                              aliases    availability_xlink      contactlist_xlink    gsf    number  street1          street2    state    city    country      zip  pobox    info
--------  ----  --------------------------------  ---------  --------------------  -------------------  -----  --------  ---------------  ---------  -------  ------  ---------  -----  -------  --------------------------------------------------------------
UPDATE     137  [-Computing Center (039)-]+AAAA+  +BBBB+     None                                  144      0       039  1225 KINCAID ST             OR       EUGENE  US         97401           === BEGIN AUTOMATION MSG ===  > WARNING: Any details entered b
''')

    

    
@pytest.mark.vcr
def test_show_changes_as_tables_select_columns(repository: Repository, capfd):
    # Arrange
    repository.enable_propose_changes(print_changes=False)
    site = repository.create_new(netdot.Site(name='Test Site'))
    floor = site.add_floor(netdot.Floor(level='Test Floor'))
    room1 = floor.add_room(netdot.Room(name='Test Room 1'))  # noqa: F841
    room2 = floor.add_room(netdot.Room(name='Test Room 2'))  # noqa: F841
    room3 = floor.add_room(netdot.Room(name='Test Room 3'))
    closet = room3.add_closet(netdot.Closet(name='Test Closet 1'))  # noqa: F841

    # Act
    repository.show_changes_as_tables(select_cols=['name', 'level'])

    # Assert
    console_output = capfd.readouterr().out

    assert_that(console_output).contains('''## Site Changes

action    id    name
--------  ----  ---------
CREATE    None  Test Site
''')
    assert_that(console_output).contains('''## Floor Changes

action    id    level
--------  ----  ----------
CREATE    None  Test Floor''')
    assert_that(console_output).contains('''## Room Changes

action    id    name
--------  ----  -----------
CREATE    None  Test Room 1
CREATE    None  Test Room 2
CREATE    None  Test Room 3''')



@pytest.mark.vcr
def test_save_changes(repository: Repository, caplog):
    # Arrange
    repository.enable_propose_changes(print_changes=False)
    site = repository.create_new(netdot.Site(name='Test Site'))
    caplog.set_level(logging.INFO)

    # Act
    repository.save_changes()

    # Assert
    assert_that(caplog.text).contains('Will CREATE Site')

    # Cleanup
    site.delete(confirm=False)


@pytest.mark.vcr
def test_status_report_when_failed_action(repository: Repository):
    # Arrange
    repository.enable_propose_changes(print_changes=False)
    site = repository.create_new(netdot.Site(name='Test Site'))
    # NOTE: Adding a 2nd site with same name will raise 400 error:
    duplicate_site = repository.create_new(netdot.Site(name='Test Site'))
    floor = site.add_floor(netdot.Floor(level='Test Floor'))
    room = floor.add_room(netdot.Room(name='Test Room'))

    netdot.defaults.ERROR_PICKLE_FILENAME = 'netdot.pickle'
    assert_that(repository.save_changes).raises(exceptions.NetdotError).when_called_with()

    # Act
    report = repository.proposed_changes.status_report()

    # Assert
    assert_that(netdot.defaults.ERROR_PICKLE_FILENAME).exists()
    assert_that(report).contains('1. Finished CREATE Site')
    assert_that(report).contains('1. Will CREATE Floor')
    assert_that(report).contains('2. Will CREATE Room')
    assert_that(report).matches(r'''Failed Action\(s\):

1. Will CREATE Site: .*
---> Failed with exception: 400 .*''')

    # Cleanup
    site.delete(confirm=False)
    os.remove(netdot.defaults.ERROR_PICKLE_FILENAME)

@pytest.mark.vcr
def test_incremental_creation_of_site_with_rooms(repository: Repository, caplog):
    # Arrange
    repository.enable_propose_changes(print_changes=False)
    site = repository.create_new(netdot.Site(name='Test Site'))
    floor = site.add_floor(netdot.Floor(level='Test Floor'))
    room1 = floor.add_room(netdot.Room(name='Test Room 1'))
    room2 = floor.add_room(netdot.Room(name='Test Room 2'))
    room3 = floor.add_room(netdot.Room(name='Test Room 3'))
    closet = room3.add_closet(netdot.Closet(name='Test Closet 1'))
    caplog.set_level(logging.INFO)

    # Act
    repository.save_changes()

    # Assert
    assert_that(site.id).is_not_none()
    assert_that(floor.id).is_not_none()
    assert_that(room1.id).is_not_none()
    assert_that(room2.id).is_not_none()
    assert_that(room3.id).is_not_none()
    assert_that(closet.id).is_not_none()

    # Cleanup
    site.delete(confirm=False)
